<?php

namespace Drupal\tone\Plugin\ToneAttachmentStrategy;

use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\tone\AttachmentStoreInterface;
use Drupal\tone\AttachmentStrategyPluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the tone_attachment_strategy.
 *
 * @ToneAttachmentStrategy(
 *   id = "tone_css_public_files",
 *   label = @Translation("CSS Public Files"),
 *   description = @Translation("Stores a stylesheet in the public files directory and adds a link tag to the HTML head.")
 * )
 */
class CssPublicFiles extends AttachmentStrategyPluginBase implements AttachmentStoreInterface, ContainerFactoryPluginInterface {

  /**
   * The file url generator service.
   */
  protected FileUrlGeneratorInterface $fileUrlGenerator;

  /**
   * The file system service.
   */
  protected FileSystemInterface $fileSystem;

  /**
   * Constructs a new CSS public files attachment strategy.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator
   *   The file url generator service.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system service.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, FileUrlGeneratorInterface $fileUrlGenerator, FileSystemInterface $fileSystem) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->fileUrlGenerator = $fileUrlGenerator;
    $this->fileSystem = $fileSystem;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator */
    $fileUrlGenerator = $container->get('file_url_generator');

    /** @var \Drupal\Core\File\FileSystemInterface $fileSystem */
    $fileSystem = $container->get('file_system');

    return new static($configuration, $plugin_id, $plugin_definition, $fileUrlGenerator, $fileSystem);
  }

  /**
   * {@inheritdoc}
   */
  public function attach(array $build, string $identifier, string $source): array {
    // Cache busting generated from hashed source.
    $b64hash = base64_encode(hash('fnv1a64', $source, TRUE));
    $bust = str_replace(['+', '/', '='], ['-', '_', ''], $b64hash);

    $uri = $this->generateUri($identifier);
    $build['#attached']['html_head_link'][][] = [
      'rel' => 'stylesheet',
      'media' => 'all',
      'href' => $this->fileUrlGenerator->generateString($uri) . '?' . $bust,
    ];

    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function put(string $identifier, string $source): bool {
    $uri = $this->generateUri($identifier);
    $path = dirname($uri);

    // Create the asset file.
    $this->fileSystem->prepareDirectory($path, FileSystemInterface::CREATE_DIRECTORY);
    try {
      if (!$this->fileSystem->saveData($source, $uri, FileSystemInterface::EXISTS_REPLACE)) {
        return FALSE;
      }
    }
    catch (FileException $e) {
      return FALSE;
    }
    // If CSS/JS gzip compression is enabled and the zlib extension is available
    // then create a gzipped version of this file. This file is served
    // conditionally to browsers that accept gzip using .htaccess rules.
    // It's possible that the rewrite rules in .htaccess aren't working on this
    // server, but there's no harm (other than the time spent generating the
    // file) in generating the file anyway. Sites on servers where rewrite rules
    // aren't working can set css.gzip to FALSE in order to skip
    // generating a file that won't be used.
    if (extension_loaded('zlib') && \Drupal::config('system.performance')->get('css.gzip')) {
      try {
        if (!$this->fileSystem->saveData(gzencode($source, 9, FORCE_GZIP), $uri . '.gz', FileSystemInterface::EXISTS_REPLACE)) {
          return FALSE;
        }
      }
      catch (FileException $e) {
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(string $identifier): bool {
    $uri = $this->generateUri($identifier);

    $result = TRUE;
    $result &= @$this->fileSystem->unlink($uri);
    @$this->fileSystem->unlink($uri . '.gz');

    return $result;
  }

  /**
   * Returns the asset uri.
   */
  protected function generateUri(string $identifier): string {
    return "public://tone_css/tone-{$identifier}.css";
  }

}
