<?php

namespace Drupal\Tests\tone\Kernel;

use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\entity_test\Entity\EntityTest;
use Drupal\entity_test\Entity\EntityTestRev;
use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\field\Traits\EntityReferenceTestTrait;
use Drupal\Tests\tone\Traits\ToneTestTrait;
use Drupal\tone\Entity\Tone;
use Drupal\tone\ToneExtraFields;
use Drupal\tone\ToneInterface;

/**
 * @coversDefaultClass \Drupal\tone\ToneExtraFields
 *
 * @group tone
 */
class ToneExtraFieldsTest extends KernelTestBase {

  use EntityReferenceTestTrait;
  use ToneTestTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'entity_test',
    'field',
    'tone',
    'user',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->installEntitySchema('tone');
    $this->installEntitySchema('entity_test');
    $this->installEntitySchema('entity_test_rev');
  }

  /**
   * @covers ::getExtraFieldInfo
   * @covers ::getFieldMap
   * @covers ::getExtraFieldDefinition
   * @covers ::getReverseReferenceGraph
   * @covers ::getAllBundlesForEntityType
   */
  public function testGetExtraFieldsInfoEmpty(): void {
    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager */
    $entityFieldManager = $this->container->get('entity_field.manager');
    /** @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entityTypeBundleInfo*/
    $entityTypeBundleInfo = $this->container->get('entity_type.bundle.info');
    /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entityDisplayRepository */
    $entityDisplayRepository = $this->container->get('entity_display.repository');
    $extraFields = new ToneExtraFields($entityFieldManager, $entityTypeBundleInfo, $entityDisplayRepository, 2);

    $extraFieldInfo = $extraFields->getExtraFieldInfo();
    $this->assertEmpty($extraFieldInfo);
  }

  /**
   * @covers ::getExtraFieldInfo
   * @covers ::getFieldMap
   * @covers ::getExtraFieldDefinition
   * @covers ::getReverseReferenceGraph
   * @covers ::getAllBundlesForEntityType
   */
  public function testGetExtraFieldsInfoDirectReference(): void {
    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager */
    $entityFieldManager = $this->container->get('entity_field.manager');
    /** @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entityTypeBundleInfo*/
    $entityTypeBundleInfo = $this->container->get('entity_type.bundle.info');
    /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entityDisplayRepository */
    $entityDisplayRepository = $this->container->get('entity_display.repository');
    $extraFields = new ToneExtraFields($entityFieldManager, $entityTypeBundleInfo, $entityDisplayRepository, 2);

    $fieldLabel = $this->randomMachineName();
    $fieldName = 'field_' . strtolower($this->randomMachineName());

    $this->createToneType();
    $this->createEntityReferenceField('entity_test', 'entity_test', $fieldName, $fieldLabel, 'tone');

    $extraFieldInfo = $extraFields->getExtraFieldInfo();
    $toneExtraField = $extraFieldInfo['entity_test']['entity_test']['display']['tone__' . $fieldName];
    $this->assertIsArray($toneExtraField);
    $this->assertInstanceOf(TranslatableMarkup::class, $toneExtraField['label']);
    $this->assertEquals('Tone from ' . $fieldLabel, (string) $toneExtraField['label']);
    $this->assertInstanceOf(TranslatableMarkup::class, $toneExtraField['description']);
    $this->assertEquals(0, $toneExtraField['weight']);
    $this->assertFalse($toneExtraField['visible']);
  }

  /**
   * @covers ::getExtraFieldInfo
   * @covers ::getFieldMap
   * @covers ::getExtraFieldDefinition
   * @covers ::getReverseReferenceGraph
   * @covers ::getAllBundlesForEntityType
   */
  public function testGetExtraFieldsInfoRecursiveReference(): void {
    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager */
    $entityFieldManager = $this->container->get('entity_field.manager');
    /** @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entityTypeBundleInfo*/
    $entityTypeBundleInfo = $this->container->get('entity_type.bundle.info');
    /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entityDisplayRepository */
    $entityDisplayRepository = $this->container->get('entity_display.repository');
    $extraFields = new ToneExtraFields($entityFieldManager, $entityTypeBundleInfo, $entityDisplayRepository, 2);

    $fieldLabel = $this->randomMachineName();
    $fieldName = 'field_' . strtolower($this->randomMachineName());
    $revLabel = $this->randomMachineName();
    $revName = 'field_' . strtolower($this->randomMachineName());

    $this->createToneType();
    $this->createEntityReferenceField('entity_test_rev', 'entity_test_rev', $fieldName, $fieldLabel, 'tone');
    $this->createEntityReferenceField('entity_test', 'entity_test', $revName, $revLabel, 'entity_test_rev');

    $extraFieldInfo = $extraFields->getExtraFieldInfo();
    $toneExtraField = $extraFieldInfo['entity_test']['entity_test']['display']['tone__' . $revName];
    $this->assertIsArray($toneExtraField);
    $this->assertInstanceOf(TranslatableMarkup::class, $toneExtraField['label']);
    $this->assertEquals('Tones from ' . $revLabel . ' (recursive)', (string) $toneExtraField['label']);
    $this->assertInstanceOf(TranslatableMarkup::class, $toneExtraField['description']);
    $this->assertEquals(0, $toneExtraField['weight']);
    $this->assertFalse($toneExtraField['visible']);
  }

  /**
   * @covers ::attach
   * @covers ::getFieldMap
   */
  public function testAttachDirectReference(): void {
    $selector = $this->randomMachineName();
    $propertyName = $this->randomMachineName();
    $fieldName = 'label';

    $bundle = $this->createToneType(
      selector: $selector,
      renderer: 'tone_css_property',
      rendererConfiguration: [
        'property_name' => $propertyName,
        'field_name' => $fieldName,
      ],
    );

    $fieldLabel = $this->randomMachineName();
    $fieldName = 'field_' . strtolower($this->randomMachineName());

    $this->createEntityReferenceField('entity_test', 'entity_test', $fieldName, $fieldLabel, 'tone');

    /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $displayRepository */
    $displayRepository = $this->container->get('entity_display.repository');

    $displayRepository->getViewDisplay('entity_test', 'entity_test')
      // Display entity reference field in default view mode.
      ->setComponent($fieldName, [
        'type' => 'entity_reference_entity_view',
        'label' => 'hidden',
        'settings' => ['view_mode' => 'full'],
      ])
      // Display tone extra field in default view mode.
      ->setComponent('tone__' . $fieldName, ['region' => 'content'])
      ->save();

    $fieldValue = $this->randomMachineName();
    $tone = Tone::create([
      'label' => $fieldValue,
      'bundle' => $bundle->id(),
    ]);
    $tone->save();

    $parentEntity = EntityTest::create([$fieldName => $tone]);
    $parentEntity->save();

    /** @var \Drupal\Core\Entity\EntityViewBuilderInterface */
    $viewBuilder = $this->container->get('entity_type.manager')->getViewBuilder('entity_test');
    $build = $viewBuilder->view($parentEntity);

    /** @var \Drupal\Core\Render\RendererInterface */
    $renderer = $this->container->get('renderer');
    $renderer->renderPlain($build);

    $this->assertCssPublicFilesAttachment($tone, $build, 'full');
  }

  /**
   * @covers ::attach
   * @covers ::getFieldMap
   */
  public function testAttachRecursiveReference(): void {
    $selector = $this->randomMachineName();
    $propertyName = $this->randomMachineName();
    $fieldName = 'label';

    $bundle = $this->createToneType(
      selector: $selector,
      renderer: 'tone_css_property',
      rendererConfiguration: [
        'property_name' => $propertyName,
        'field_name' => $fieldName,
      ],
    );

    $fieldLabel = $this->randomMachineName();
    $fieldName = 'field_' . strtolower($this->randomMachineName());
    $revLabel = $this->randomMachineName();
    $revName = 'field_' . strtolower($this->randomMachineName());

    $this->createEntityReferenceField('entity_test_rev', 'entity_test_rev', $fieldName, $fieldLabel, 'tone');
    $this->createEntityReferenceField('entity_test', 'entity_test', $revName, $revLabel, 'entity_test_rev');

    /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $displayRepository */
    $displayRepository = $this->container->get('entity_display.repository');

    $displayRepository->getViewDisplay('entity_test_rev', 'entity_test_rev')
      // Display entity reference field in default view mode.
      ->setComponent($fieldName, [
        'type' => 'entity_reference_entity_view',
        'label' => 'hidden',
        'settings' => ['view_mode' => 'full'],
      ])
      // Display tone extra field in default view mode.
      ->setComponent('tone__' . $fieldName, ['region' => 'content'])
      ->save();

    $displayRepository->getViewDisplay('entity_test', 'entity_test')
      // Display entity reference field in default view mode.
      ->setComponent($revName, [
        'type' => 'entity_reference_entity_view',
        'label' => 'hidden',
        'settings' => ['view_mode' => 'default'],
      ])
      // Display tone extra field in default view mode.
      ->setComponent('tone__' . $revName, ['region' => 'content'])
      ->save();

    $fieldValue = $this->randomMachineName();
    $tone = Tone::create([
      'label' => $fieldValue,
      'bundle' => $bundle->id(),
    ]);
    $tone->save();

    $parentEntity = EntityTestRev::create([$fieldName => $tone]);
    $parentEntity->save();

    $topEntity = EntityTest::create([$revName => $parentEntity]);
    $topEntity->save();

    /** @var \Drupal\Core\Entity\EntityViewBuilderInterface */
    $viewBuilder = $this->container->get('entity_type.manager')->getViewBuilder('entity_test');
    $build = $viewBuilder->view($topEntity);

    /** @var \Drupal\Core\Render\RendererInterface */
    $renderer = $this->container->get('renderer');
    $renderer->renderPlain($build);

    $this->assertCssPublicFilesAttachment($tone, $build, 'full');
  }

  /**
   * Asserts that a HTML head link to a tone css file is attached.
   */
  protected function assertCssPublicFilesAttachment(ToneInterface $tone, array $build) {
    /** @var \Drupal\tone\IdentityInterface $identity */
    $identity = $tone->bundle->entity->getIdentity();
    $identifier = $identity->getIdentifier($tone, $build['#view_mode']);

    /** @var \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator */
    $fileUrlGenerator = $this->container->get('file_url_generator');
    $expectedUrlPrefix = $fileUrlGenerator->generateString("public://tone_css/tone-{$identifier}.css");

    $url = $build['#attached']['html_head_link'][0][0]['href'];
    $this->assertStringStartsWith($expectedUrlPrefix, $url);
  }

}
